package ru.dbotthepony.mc.otm.android

import com.google.common.collect.ImmutableList
import com.google.common.collect.Streams
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSyntaxException
import com.google.gson.internal.bind.TypeAdapters
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.ComponentContents
import net.minecraft.network.chat.MutableComponent
import net.minecraft.network.chat.contents.TranslatableContents
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import net.minecraft.world.item.crafting.Ingredient
import net.minecraft.world.level.ItemLike
import net.minecraftforge.registries.ForgeRegistries
import ru.dbotthepony.mc.otm.client.render.AbstractMatterySprite
import ru.dbotthepony.mc.otm.client.render.SpriteType
import ru.dbotthepony.mc.otm.core.collect.ListSet
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.isActuallyEmpty
import ru.dbotthepony.mc.otm.core.registryName
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.core.toImmutableList
import ru.dbotthepony.mc.otm.core.collect.stream
import ru.dbotthepony.mc.otm.isClient
import ru.dbotthepony.mc.otm.registry.MRegistry
import java.util.LinkedList
import java.util.stream.Stream
import kotlin.collections.ArrayList
import kotlin.collections.HashSet

private fun findPrerequisites(
	initial: Collection<AndroidResearchType>,
	add: MutableSet<AndroidResearchType> = HashSet(),
	top: Boolean = true
): Set<AndroidResearchType> {
	for (value in initial) {
		if (!top) {
			add.add(value)
		}

		findPrerequisites(value.resolvedPrerequisites, add, false)
	}

	return add
}

private fun findAllPrerequisites(
	initial: Collection<AndroidResearchType>,
	add: MutableSet<AndroidResearchType> = HashSet(),
): Set<AndroidResearchType> {
	for (value in initial) {
		add.add(value)
		findAllPrerequisites(value.flatPrerequisites, add)
	}

	return add
}

private fun findAllChildren(
	initial: Collection<AndroidResearchType>,
	add: MutableSet<AndroidResearchType> = HashSet(),
): Set<AndroidResearchType> {
	for (value in initial) {
		add.add(value)
		findAllChildren(value.flatUnlocks, add)
	}

	return add
}

class AndroidResearchType(
	val id: ResourceLocation,
	prerequisites: Collection<Reference>,
	blockedBy: Collection<Reference>,

	items: Collection<Pair<Ingredient, Int>>,
	features: Collection<FeatureReference>,

	descriptionLines: Collection<Component>,
	descriptionSuppliers: Collection<ComponentSupplier> = listOf(),

	val experienceLevels: Int = 0,
	private val customName: Component? = null,

	// ok
	val skinIcon: JsonObject? = null,
	val itemIcon: Item? = null,
	iconText: Component? = null,
) {
	private val iconTextValue = iconText?.copy()
	val iconText get() = iconTextValue?.copy()

	val resolvedSkinIcon by lazy {
		check(isClient) { "Invalid realm" }
		if (skinIcon != null)
			SpriteType.fromJson(skinIcon)
		else
			null
	}

	data class Reference(
		val id: ResourceLocation,
		val isRigid: Boolean
	) {
		fun toJson(): JsonObject {
			return JsonObject().also {
				it["id"] = JsonPrimitive(id.toString())
				it["is_rigid"] = JsonPrimitive(isRigid)
			}
		}

		fun toNetwork(buff: FriendlyByteBuf) {
			buff.writeUtf(id.toString())
			buff.writeBoolean(isRigid)
		}

		companion object {
			fun fromNetwork(buff: FriendlyByteBuf): Reference {
				return Reference(
					ResourceLocation(buff.readUtf()),
					buff.readBoolean()
				)
			}

			fun fromJson(value: JsonElement): Reference {
				if (value is JsonPrimitive) {
					return Reference(ResourceLocation(value.asString), true)
				} else if (value is JsonObject) {
					return Reference(
						ResourceLocation((value["id"] as? JsonPrimitive ?: throw JsonSyntaxException("Invalid `id` value")).asString),
						(value["is_rigid"] as JsonPrimitive?)?.asBoolean ?: true)
				} else {
					throw JsonSyntaxException("Unknown element type ${value::class.qualifiedName}")
				}
			}
		}
	}

	data class FeatureReference(
		val id: ResourceLocation,
		val level: Int = 0,
		val isRigid: Boolean,
		val transformersUp: Collection<AndroidResultTransformer> = listOf(),
		val transformersDown: Collection<AndroidResultTransformer> = listOf(),
	) {
		fun toJson(): JsonObject {
			return JsonObject().also {
				it["id"] = JsonPrimitive(id.toString())
				it["level"] = JsonPrimitive(level)
				it["is_rigid"] = JsonPrimitive(isRigid)
				it["functions_up"] = JsonArray().also {
					for (transformer in transformersUp) {
						it.add(transformer.toJson())
					}
				}

				it["functions_down"] = JsonArray().also {
					for (transformer in transformersDown) {
						it.add(transformer.toJson())
					}
				}
			}
		}

		fun toNetwork(buff: FriendlyByteBuf) {
			buff.writeUtf(id.toString())
			buff.writeVarInt(level)
			buff.writeBoolean(isRigid)
			buff.writeCollection(transformersUp) { a, b -> b.toNetwork(a) }
			buff.writeCollection(transformersDown) { a, b -> b.toNetwork(a) }
		}

		companion object {
			fun fromNetwork(buff: FriendlyByteBuf): FeatureReference {
				return FeatureReference(
					ResourceLocation(buff.readUtf()),
					buff.readVarInt(),
					buff.readBoolean(),
					buff.readCollection({ LinkedList() }, AndroidResearchManager.featureResultTransformers::fromNetwork).filterNotNull().toImmutableList(),
					buff.readCollection({ LinkedList() }, AndroidResearchManager.featureResultTransformers::fromNetwork).filterNotNull().toImmutableList()
				)
			}

			fun fromJson(value: JsonElement): FeatureReference {
				if (value is JsonPrimitive) {
					return FeatureReference(ResourceLocation(value.asString), 0, true)
				} else if (value is JsonObject) {
					return FeatureReference(
						ResourceLocation((value["id"] as? JsonPrimitive ?: throw JsonSyntaxException("Invalid `id` value")).asString),
						(value["level"] as JsonPrimitive?)?.asInt ?: 0,
						(value["is_rigid"] as JsonPrimitive?)?.asBoolean ?: true,
						(value["functions_up"] as JsonArray? ?: JsonArray()).stream().map { AndroidResearchManager.featureResultTransformers.fromJson(it) }.filter { it != null }.collect(ImmutableList.toImmutableList<AndroidResultTransformer>()),
						(value["functions_down"] as JsonArray? ?: JsonArray()).stream().map { AndroidResearchManager.featureResultTransformers.fromJson(it) }.filter { it != null }.collect(ImmutableList.toImmutableList<AndroidResultTransformer>()),
					)
				} else {
					throw JsonSyntaxException("Unknown element type ${value::class.qualifiedName}")
				}
			}
		}
	}

	data class ResolvedFeature(
		val feature: AndroidFeatureType<*>,
		val level: Int,
	)

	val researchTreeDepth: Int by lazy {
		if (flatPrerequisites.isEmpty()) {
			return@lazy 0
		}

		var depth = 1

		for (value in flatPrerequisites) {
			depth = depth.coerceAtLeast(value.researchTreeDepth + 1)
		}

		return@lazy depth
	}

	/**
	 * Prerequisites as-is.
	 *
	 * Please avoid having more than one prerequisite, as this case doesn't have proper research tree
	 * rendering code (yet).
	 */
	val prerequisites: List<Reference> = ImmutableList.copyOf(prerequisites)

	/**
	 * Blocked by list as-is
	 */
	val blockedBy: List<Reference> = ImmutableList.copyOf(blockedBy)

	val resolvedPrerequisites: List<AndroidResearchType> by lazy {
		ImmutableList.copyOf(this.prerequisites.mapNotNull { AndroidResearchManager[it.id].also { e -> if (e == null && it.isRigid) throw NoSuchElementException("Unable to find research ${it.id}") } })
	}

	val resolvedBlockedBy: List<AndroidResearchType> by lazy {
		ImmutableList.copyOf(this.blockedBy.mapNotNull { AndroidResearchManager[it.id].also { e -> if (e == null && it.isRigid) throw NoSuchElementException("Unable to find research ${it.id}") } })
	}

	val features: List<FeatureReference> = ImmutableList.copyOf(features)

	val resolvedFeatures: List<ResolvedFeature> by lazy {
		ImmutableList.copyOf(features.mapNotNull {
			MRegistry.ANDROID_FEATURES.getValue(it.id)
				.let { e -> if (e == null && it.isRigid)
							throw NoSuchElementException("Unable to find research ${it.id}")
						else if (e != null)
							ResolvedFeature(e, it.level)
						else
							null } })
	}

	private val definedItems: List<Pair<Ingredient, Int>> = ImmutableList.copyOf(items)

	val items: Stream<out Pair<Ingredient, Int>> get() = definedItems.stream().filter { !it.first.isActuallyEmpty }

	private val descriptionLines: List<MutableComponent> = ImmutableList.copyOf(descriptionLines.map { it.copy() })

	private val descriptionSuppliers: List<ComponentSupplier> = ImmutableList.copyOf(descriptionSuppliers)

	/**
	 * Stream containing copies of original [Component]s in list
	 */
	val description: Stream<out Component> get() = Streams.concat(descriptionLines.stream().map { it.copy() }, descriptionSuppliers.stream().map { it.apply(Unit) })

	/**
	 * Flat list of research preceding this research.
	 *
	 * This list won't contain research following the same research path twice (guaranteed to be flat).
	 *
	 * E.g.
	 *
	 * * C depends on B
	 * * B depends on A
	 *
	 * C specify both B and A as it's prerequisites, [flatPrerequisites] will contain only B, when [prerequisites] will contain
	 * both B and A
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val flatPrerequisites: List<AndroidResearchType> by lazy {
		val parentPrerequisites = findPrerequisites(resolvedPrerequisites)
		val builder = ImmutableList.builder<AndroidResearchType>()

		for (value in resolvedPrerequisites) {
			if (value !in parentPrerequisites) {
				builder.add(value)
			}
		}

		ListSet(builder.build())
	}

	/**
	 * All prerequisite research, including indirect ones
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val allPrerequisites: List<AndroidResearchType> by lazy {
		ListSet(findAllPrerequisites(flatPrerequisites))
	}

	/**
	 * Flat list of research blocked by this research.
	 *
	 * This list won't contain children of any blocked research (guaranteed to be flat).
	 *
	 * E.g.
	 *
	 * * C depends on B
	 * * B is blocked by A
	 * * C is blocked by A
	 *
	 * Both C and B specify A as it's blockers, [flatBlocking] will contain only B, because C is blocked by A through B.
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val flatBlocking: List<AndroidResearchType> by lazy {
		val list = ImmutableList.builder<AndroidResearchType>()

		for (research in AndroidResearchManager.researchMap.values) {
			if (this in research.resolvedBlockedBy) {
				var hit = false

				for (parent in research.allPrerequisites) {
					if (this in parent.resolvedBlockedBy) {
						hit = true
						break
					}
				}

				if (!hit) {
					list.add(research)
				}
			}
		}

		ListSet(list.build())
	}

	/**
	 * Flat list of research blocking this research.
	 *
	 * This list won't contain any research blocked by this' research parents.
	 *
	 * E.g.
	 *
	 * * C depends on B
	 * * B is blocked by A
	 * * C is blocked by A
	 *
	 * Both B and C are blocked by A, but C requires B, so C's [flatBlockedBy] will not contain A, but B's will.
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val flatBlockedBy: ListSet<AndroidResearchType> by lazy {
		val list = ImmutableList.builder<AndroidResearchType>()

		for (blocker in resolvedBlockedBy) {
			var hit = false

			for (research in allPrerequisites) {
				if (blocker in research.flatBlockedBy) {
					hit = true
					break
				}
			}

			if (!hit) {
				list.add(blocker)
			}
		}

		ListSet(list.build())
	}

	/**
	 * All research blocking this research, even indirect ones (e.g. blocked by through one of parent research).
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val allBlockedBy: ListSet<AndroidResearchType> by lazy {
		val list = HashSet<AndroidResearchType>()

		list.addAll(flatBlockedBy)

		for (parent in allPrerequisites) {
			list.addAll(parent.flatBlockedBy)
		}

		ListSet(list)
	}

	/**
	 * All research directly unlocked by this research.
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val flatUnlocks: List<AndroidResearchType> by lazy {
		val list = ImmutableList.builder<AndroidResearchType>()

		for (research in AndroidResearchManager.researchMap.values) {
			if (this in research.flatPrerequisites) {
				list.add(research)
			}
		}

		ListSet(list.build())
	}

	/**
	 * All research unlocked by this research, including indirect ones
	 *
	 * Returns list which also doubles as set (for contains method).
	 */
	val allUnlocks: List<AndroidResearchType> by lazy {
		ListSet(findAllChildren(flatUnlocks))
	}

	/**
	 * All research blocked by this research, including indirect ones.
	 *
	 * Returns list which also doubles as set (for contains method)
	 */
	val allBlocking: List<AndroidResearchType> by lazy {
		val set = HashSet<AndroidResearchType>()

		for (research in flatBlocking) {
			set.add(research)
			set.addAll(research.allUnlocks)
		}

		for (research in allPrerequisites) {
			for (research2 in research.flatBlocking) {
				set.add(research2)
				set.addAll(research2.allUnlocks)
			}
		}

		ListSet(set)
	}

	val displayId by lazy {
		return@lazy "android_research.${id.namespace}.${id.path}".intern()
	}

	val descriptionId by lazy {
		return@lazy "$displayId.description".intern()
	}

	val displayContents: ComponentContents by lazy {
		TranslatableContents(displayId, null, arrayOf())
	}

	val displayName: Component get() {
		return customName?.copy() ?: MutableComponent.create(displayContents)
	}

	fun toJson(): JsonElement {
		return JsonObject().also {
			// it["id"] = JsonPrimitive(id.toString())

			it["prerequisites"] = JsonArray().also { for (value in prerequisites) it.add(value.toJson()) }
			it["blocked_by"] = JsonArray().also { for (value in blockedBy) it.add(value.toJson()) }
			it["required_items"] = JsonArray().also { for (item in definedItems) it.add(JsonObject().also { it["count"] = JsonPrimitive(item.second); it["ingredient"] = item.first.toJson() }) }
			it["feature_result"] = JsonArray().also { for (feature in features) it.add(feature.toJson()) }
			it["description"] = JsonArray().also { for (line in descriptionLines) it.add(Component.Serializer.toJsonTree(line)) }
			it["description_funcs"] = JsonArray().also { for (line in descriptionSuppliers) it.add(line.toJson()) }
			it["experience"] = JsonPrimitive(experienceLevels)

			if (skinIcon != null) {
				it["skin_icon"] = skinIcon
			}

			if (itemIcon != null) {
				it["item_icon"] = JsonPrimitive(itemIcon.registryName!!.toString())
			}

			if (iconTextValue != null) {
				it["icon_text"] = Component.Serializer.toJsonTree(iconTextValue)
			}

			if (customName != null) {
				it["custom_name"] = Component.Serializer.toJsonTree(customName)
			}
		}
	}

	fun validate() {
		resolvedFeatures
		resolvedBlockedBy
		resolvedPrerequisites
	}

	fun toNetwork(buff: FriendlyByteBuf) {
		buff.writeUtf(id.toString())
		buff.writeCollection(prerequisites) { a, b -> b.toNetwork(a) }
		buff.writeCollection(blockedBy) { a, b -> b.toNetwork(a) }
		buff.writeCollection(definedItems) { a, b -> b.first.toNetwork(a); a.writeVarInt(b.second) }
		buff.writeCollection(features) { a, b -> b.toNetwork(a) }
		buff.writeCollection(descriptionLines) { a, b -> a.writeComponent(b) }
		buff.writeCollection(descriptionSuppliers) { a, b -> b.toNetwork(a) }
		buff.writeVarInt(experienceLevels)

		buff.writeBoolean(customName != null)
		if (customName != null) buff.writeComponent(customName)

		buff.writeBoolean(iconTextValue != null)
		if (iconTextValue != null) buff.writeComponent(iconTextValue)

		buff.writeBoolean(skinIcon != null)
		if (skinIcon != null) buff.writeUtf(skinIcon.toString())

		buff.writeBoolean(itemIcon != null)
		if (itemIcon != null) buff.writeUtf(itemIcon.registryName!!.toString())
	}

	companion object {
		fun fromNetwork(buff: FriendlyByteBuf): AndroidResearchType {
			val id = ResourceLocation(buff.readUtf())
			val prerequisites = buff.readCollection({ LinkedList() }, Reference::fromNetwork)
			val blockedBy = buff.readCollection({ LinkedList() }, Reference::fromNetwork)
			val items = buff.readCollection({ LinkedList() }, { Ingredient.fromNetwork(it) to it.readVarInt() })
			val features = buff.readCollection({ LinkedList() }, FeatureReference::fromNetwork)
			val descriptionLines = buff.readCollection({ LinkedList() }, FriendlyByteBuf::readComponent)
			val descriptionSuppliers = buff.readCollection({ LinkedList() }, { AndroidResearchManager.descriptionFuncs.fromNetwork(it) })
			val experienceLevels = buff.readVarInt()

			val customName = if (buff.readBoolean()) {
				buff.readComponent()
			} else {
				null
			}

			val iconTextValue = if (buff.readBoolean()) {
				buff.readComponent()
			} else {
				null
			}

			val skinIcon = if (buff.readBoolean()) {
				TypeAdapters.JSON_ELEMENT.fromJson(buff.readUtf()) as JsonObject
			} else {
				null
			}

			val itemIcon = if (buff.readBoolean()) {
				ForgeRegistries.ITEMS.getValue(ResourceLocation(buff.readUtf()))?.let { if (it == Items.AIR) null else it }
			} else {
				null
			}

			return AndroidResearchType(
				id = id,
				prerequisites = prerequisites,
				blockedBy = blockedBy,
				items = items,
				features = features,
				descriptionLines = descriptionLines,
				descriptionSuppliers = descriptionSuppliers.filterNotNull(),
				experienceLevels = experienceLevels,
				customName = customName,
				iconText = iconTextValue,
				skinIcon = skinIcon,
				itemIcon = itemIcon,
			)
		}

		fun fromJson(value: JsonElement, id: ResourceLocation): AndroidResearchType {
			if (value !is JsonObject) {
				throw JsonSyntaxException("Android research type must be of Json Object")
			}

			val prerequisites = value["prerequisites"] as JsonArray? ?: JsonArray()
			val blocked_by = value["blocked_by"] as JsonArray? ?: JsonArray()
			val items = value["required_items"] as JsonArray? ?: JsonArray()
			val features = value["feature_result"] as JsonArray? ?: JsonArray()
			val description = value["description"] as JsonArray? ?: JsonArray()
			val description_funcs = value["description_funcs"] as JsonArray? ?: JsonArray()
			val experience = value["experience"]?.asInt ?: 0
			val customName = value["custom_name"]?.let(Component.Serializer::fromJson)
			val iconText = value["icon_text"]?.let(Component.Serializer::fromJson)
			val skinIcon = value["skin_icon"] as JsonObject?
			val itemIcon = value["item_icon"]?.let { ForgeRegistries.ITEMS.getValue(ResourceLocation(it.asString)).let { if (it == Items.AIR) null else it } }

			return AndroidResearchType(
				id = id,
				prerequisites = prerequisites.stream().map { Reference.fromJson(it) }.toList(),
				blockedBy = blocked_by.stream().map { Reference.fromJson(it) }.toList(),
				features = features.stream().map { FeatureReference.fromJson(it) }.toList(),
				items = items.stream()
					.map { it as? JsonObject ?: throw JsonSyntaxException("One of items is not an JsonObject") }
					.map { Ingredient.fromJson(it["ingredient"] ?: throw JsonSyntaxException("Missing ingredient key")) to (it["count"]?.asInt ?: throw JsonSyntaxException("Missing count key")) }
					.toList(),
				descriptionLines = description.stream().map { Component.Serializer.fromJson(it) }.toList() as List<MutableComponent>,
				descriptionSuppliers = description_funcs.stream()
					.map { AndroidResearchManager.descriptionFuncs.fromJson(it) ?: throw NullPointerException("$id is missing description supplier function or it is invalid! JSON: $it") }
					.toList() as List<ComponentSupplier>,
				experienceLevels = experience,
				customName = customName,
				iconText = iconText,
				skinIcon = skinIcon,
				itemIcon = itemIcon,
			)
		}
	}

	@Suppress("unused")
	class Builder(
		val id: ResourceLocation,
		var experience: Int = 0,
		var customName: Component? = null,
		var description: MutableList<Component>? = null,
		var descriptionSuppliers: MutableList<ComponentSupplier>? = null,
		var itemIcon: Item? = null,
		var skinIcon: AbstractMatterySprite? = null,
		var iconText: Component? = null,
	) {
		private val items = ArrayList<Pair<Ingredient, Int>>()
		private val prerequisites = ArrayList<Reference>()
		private val blockers = ArrayList<Reference>()

		private val features = ArrayList<FeatureReference>()

		fun withIconText(icon: Component? = null): Builder {
			this.iconText = icon
			return this
		}

		fun withIcon(icon: AbstractMatterySprite? = null): Builder {
			this.skinIcon = icon
			this.itemIcon = null
			return this
		}

		fun withIcon(icon: Item? = null): Builder {
			this.itemIcon = icon
			this.skinIcon = null
			return this
		}

		fun withIcon(): Builder {
			this.itemIcon = null
			this.skinIcon = null
			return this
		}

		fun withName(customName: Component): Builder {
			this.customName = customName
			return this
		}

		fun withDescription(): Builder {
			this.description = mutableListOf(TranslatableComponent("android_research.${id.namespace}.${id.path}.description"))
			return this
		}

		fun withDescription(range: IntRange): Builder {
			val result = ArrayList<Component>()

			for (i in range) {
				result.add(TranslatableComponent("android_research.${id.namespace}.${id.path}.description$i"))
			}

			this.description = result
			return this
		}

		fun withExperience(experience: Int): Builder {
			this.experience = experience
			return this
		}

		fun withDescription(vararg description: Component): Builder {
			this.description = description.toMutableList()
			return this
		}

		fun withDescription(description: Collection<Component>): Builder {
			this.description = ArrayList<Component>(description.size).also { it.addAll(description) }
			return this
		}

		fun appendDescription(range: IntRange): Builder {
			val result = this.description ?: ArrayList()

			for (i in range) {
				result.add(TranslatableComponent("android_research.${id.namespace}.${id.path}.description$i"))
			}

			this.description = result
			return this
		}

		fun appendDescription(description: Component): Builder {
			this.description = (this.description ?: mutableListOf()).also { it.add(description) }
			return this
		}

		fun appendDescription(vararg description: Component): Builder {
			this.description = (this.description ?: mutableListOf()).also { it.addAll(description) }
			return this
		}

		fun appendDescription(description: Collection<Component>): Builder {
			this.description = (this.description ?: mutableListOf()).also { it.addAll(description) }
			return this
		}

		fun withDescription(vararg description: ComponentSupplier): Builder {
			this.descriptionSuppliers = description.toMutableList()
			return this
		}

		fun withDescriptionSupplier(description: Collection<ComponentSupplier>): Builder {
			this.descriptionSuppliers = ArrayList<ComponentSupplier>(description.size).also { it.addAll(description) }
			return this
		}

		fun appendDescription(description: ComponentSupplier): Builder {
			this.descriptionSuppliers = (this.descriptionSuppliers ?: mutableListOf()).also { it.add(description) }
			return this
		}

		fun appendDescriptionSupplier(description: ComponentSupplier): Builder {
			this.descriptionSuppliers = (this.descriptionSuppliers ?: mutableListOf()).also { it.add(description) }
			return this
		}

		fun appendDescription(vararg description: ComponentSupplier): Builder {
			this.descriptionSuppliers = (this.descriptionSuppliers ?: mutableListOf()).also { it.addAll(description) }
			return this
		}

		fun appendDescriptionSupplier(vararg description: ComponentSupplier): Builder {
			this.descriptionSuppliers = (this.descriptionSuppliers ?: mutableListOf()).also { it.addAll(description) }
			return this
		}

		fun appendDescriptionSupplier(description: Collection<ComponentSupplier>): Builder {
			this.descriptionSuppliers = (this.descriptionSuppliers ?: mutableListOf()).also { it.addAll(description) }
			return this
		}

		/**
		 * Please avoid having multiple prerequisites as case with more than 1 prerequisite does not have proper
		 * research tree render logic (yet).
		 */
		@JvmOverloads
		fun addPrerequisite(id: ResourceLocation, rigid: Boolean = false): Builder {
			prerequisites.add(Reference(id, rigid))
			return this
		}

		@JvmOverloads
		fun addBlocker(id: ResourceLocation, rigid: Boolean = false): Builder {
			blockers.add(Reference(id, rigid))
			return this
		}

		/**
		 * Please avoid having multiple prerequisites as case with more than 1 prerequisite does not have proper
		 * research tree render logic (yet).
		 */
		fun addPrerequisite(type: AndroidResearchType, rigid: Boolean = true) = addPrerequisite(type.id, rigid)
		fun addBlocker(type: AndroidResearchType, rigid: Boolean = true) = addBlocker(type.id, rigid)

		@JvmOverloads
		fun addFeatureResult(
			id: ResourceLocation,
			level: Int = 0,
			rigid: Boolean = false,
			transformersUp: Collection<AndroidResultTransformer> = listOf(),
			transformersDown: Collection<AndroidResultTransformer> = listOf(),
		): Builder {
			features.add(FeatureReference(id, level, rigid, transformersUp, transformersDown))
			return this
		}

		@JvmOverloads
		fun addFeatureResult(
			feature: AndroidFeatureType<*>,
			level: Int = 0,
			rigid: Boolean = true,
			transformersUp: Collection<AndroidResultTransformer> = listOf(),
			transformersDown: Collection<AndroidResultTransformer> = listOf(),
		): Builder {
			features.add(FeatureReference(feature.registryName ?: throw NullPointerException("Feature $feature does not have registry name"), level, rigid, transformersUp, transformersDown))
			return this
		}

		fun addFeatureResult(
			id: ResourceLocation,
			rigid: Boolean = false,
			transformersUp: Collection<AndroidResultTransformer> = listOf(),
			transformersDown: Collection<AndroidResultTransformer> = listOf(),
		): Builder {
			features.add(FeatureReference(id, 0, rigid, transformersUp, transformersDown))
			return this
		}

		fun addFeatureResult(ref: FeatureReference): Builder {
			features.add(ref)
			return this
		}

		fun addItem(cost: ItemStack): Builder {
			require(!cost.isEmpty) { "Provided stack is empty" }
			items.add(Ingredient.of(cost) to cost.count)
			return this
		}

		fun addItem(cost: ItemLike, count: Int = 1): Builder {
			items.add(Ingredient.of(cost) to count)
			return this
		}

		fun addItem(cost: Ingredient, count: Int = 1): Builder {
			require(count >= 1) { "Invalid count $count" }
			items.add(cost to count)
			return this
		}

		fun addItem(cost: TagKey<Item>, count: Int = 1): Builder {
			require(count >= 1) { "Invalid count $count" }
			items.add(Ingredient.of(cost) to count)
			return this
		}

		fun build(validate: Boolean = false): AndroidResearchType {
			return AndroidResearchType(
				id = id,
				prerequisites = prerequisites,
				blockedBy = blockers,
				items = items,
				features = features,
				descriptionLines = description ?: listOf(),
				descriptionSuppliers = descriptionSuppliers ?: listOf(),
				experienceLevels = experience,
				customName = customName,
				skinIcon = skinIcon?.toJson(),
				itemIcon = itemIcon,
				iconText = iconText
			).also { if (validate) it.validate() }
		}
	}
}
