package ru.dbotthepony.mc.otm.matter

import com.google.common.collect.ImmutableList
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import com.google.gson.JsonSyntaxException
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.ItemTags
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import net.minecraft.world.level.ItemLike
import net.minecraftforge.registries.ForgeRegistries
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.registryName
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.core.collect.stream
import java.util.stream.Stream

class ComputeAction : AbstractRegistryAction {
	sealed class Source {
		val matterFunction: MatterFunction
		val complexityFunction: MatterFunction

		constructor(
			matterFunction: MatterFunction,
			complexityFunction: MatterFunction,
		) {
			this.matterFunction = matterFunction
			this.complexityFunction = complexityFunction
		}

		constructor(json: JsonObject) {
			val function = json["function"]?.asString?.uppercase()

			if (function == null) {
				try {
					this.matterFunction = MatterFunction.valueOf(json["matter_function"]?.asString?.uppercase() ?: throw JsonSyntaxException("Missing matter_function value"))
				} catch(err: NoSuchElementException) {
					throw JsonSyntaxException("Invalid matter_function: ${json["matter_function"]}", err)
				}

				try {
					this.complexityFunction = MatterFunction.valueOf(json["complexity_function"]?.asString?.uppercase() ?: throw JsonSyntaxException("Missing complexity_function value"))
				} catch(err: NoSuchElementException) {
					throw JsonSyntaxException("Invalid complexity_function: ${json["complexity_function"]}", err)
				}
			} else {
				try {
					this.matterFunction = MatterFunction.valueOf(function)
					this.complexityFunction = this.matterFunction
				} catch(err: NoSuchElementException) {
					throw JsonSyntaxException("Invalid function: ${json["function"]}", err)
				}
			}
		}

		internal abstract fun provideValues(): MatterManager.Result

		open fun toJson(): JsonObject {
			return JsonObject().also {
				it["matter_function"] = matterFunction.name
				it["complexity_function"] = complexityFunction.name
			}
		}

		internal fun update(self: IMatterValue): MatterManager.Result {
			val grab = provideValues()

			if (grab.value == null) {
				return grab
			}

			return MatterManager.Result(MatterValue(
				matterFunction.updateValue(self.matter, grab.value.matter),
				complexityFunction.updateValue(self.complexity, grab.value.complexity)
			))
		}
	}

	class Constant : Source {
		val matter: Decimal
		val complexity: Double

		constructor(
			matter: Decimal,
			complexity: Double,
			matterFunction: MatterFunction,
			complexityFunction: MatterFunction = matterFunction,
		) : super(matterFunction, complexityFunction) {
			this.matter = matter
			this.complexity = complexity
		}

		constructor(json: JsonObject) : super(json) {
			try {
				matter = Decimal(json["matter"]?.asString ?: throw JsonSyntaxException("Missing matter value"))
			} catch(err: NumberFormatException) {
				throw JsonSyntaxException("Invalid matter value: ${json["matter"]}")
			}

			try {
				complexity = json["complexity"]?.asDouble ?: throw JsonSyntaxException("Missing complexity value")
			} catch(err: NumberFormatException) {
				throw JsonSyntaxException("Invalid complexity value: ${json["complexity"]}")
			}
		}

		override fun provideValues(): MatterManager.Result {
			return MatterManager.Result(MatterValue(matter, complexity))
		}

		override fun toJson(): JsonObject {
			return super.toJson().also {
				it["type"] = "constant"
				it["matter"] = matter.toString()
				it["complexity"] = complexity
			}
		}
	}

	class Key : Source {
		val key: ResourceLocation

		constructor(
			key: ResourceLocation,
			matterFunction: MatterFunction,
			complexityFunction: MatterFunction = matterFunction,
		) : super(matterFunction, complexityFunction) {
			this.key = key
		}

		constructor(
			key: ItemLike,
			matterFunction: MatterFunction,
			complexityFunction: MatterFunction = matterFunction,
		) : super(matterFunction, complexityFunction) {
			this.key = key.asItem().registryName ?: throw IllegalStateException("${key.asItem()} has no registry ID!")
		}

		constructor(json: JsonObject) : super(json) {
			key = ResourceLocation.tryParse(json["key"]?.asString ?: throw JsonSyntaxException("Missing key")) ?: throw JsonSyntaxException("Invalid key value: ${json["key"]}")
		}

		override fun provideValues(): MatterManager.Result {
			return MatterManager.compute(ForgeRegistries.ITEMS.getValue(key) ?: throw NullPointerException("$key does not point to anything"))
		}

		override fun toJson(): JsonObject {
			return super.toJson().also {
				it["type"] = "key"
				it["key"] = key.toString()
			}
		}
	}

	class Tag : Source {
		val tag: TagKey<Item>

		constructor(
			tag: TagKey<Item>,
			matterFunction: MatterFunction,
			complexityFunction: MatterFunction = matterFunction,
		) : super(matterFunction, complexityFunction) {
			this.tag = tag
		}

		constructor(json: JsonObject) : super(json) {
			tag = ItemTags.create(ResourceLocation.tryParse(json["tag"]?.asString ?: throw JsonSyntaxException("Missing tag")) ?: throw JsonSyntaxException("Invalid tag value: ${json["tag"]}"))
		}

		override fun provideValues(): MatterManager.Result {
			return MatterManager.compute(tag)
		}

		override fun toJson(): JsonObject {
			return super.toJson().also {
				it["type"] = "tag"
				it["tag"] = tag.location.toString()
			}
		}
	}

	val values: List<Source>

	constructor(
		key: ResourceLocation,
		value: Source
	) : super(key = key, errorOnFailure = false, matter = null, complexity = null) {
		values = listOf(value)
	}

	constructor(
		tag: TagKey<Item>,
		value: Source
	) : super(tag = tag, errorOnFailure = false, matter = null, complexity = null) {
		values = listOf(value)
	}

	constructor(
		key: ResourceLocation,
		value: Stream<Source>
	) : super(key = key, errorOnFailure = false, matter = null, complexity = null) {
		values = value.collect(ImmutableList.toImmutableList())
		require(values.isNotEmpty()) { "Provided compute source Stream is empty" }
	}

	constructor(
		tag: TagKey<Item>,
		value: Stream<Source>
	) : super(tag = tag, errorOnFailure = false, matter = null, complexity = null) {
		values = value.collect(ImmutableList.toImmutableList())
		require(values.isNotEmpty()) { "Provided compute source Stream is empty" }
	}

	constructor(
		key: ResourceLocation,
		value: Collection<Source>
	) : super(key = key, errorOnFailure = false, matter = null, complexity = null) {
		require(value.isNotEmpty()) { "Provided compute source list is empty" }
		values = ImmutableList.copyOf(value)
	}

	constructor(
		tag: TagKey<Item>,
		value: Collection<Source>
	) : super(tag = tag, errorOnFailure = false, matter = null, complexity = null) {
		require(value.isNotEmpty()) { "Provided compute source list is empty" }
		values = ImmutableList.copyOf(value)
	}

	internal fun tryCompute(): MatterManager.Result {
		var compute: IMatterValue = IMatterValue.Companion

		for (fn in values) {
			val result = fn.update(compute)

			if (result.value == null) {
				return result
			} else {
				compute = result.value
			}
		}

		return MatterManager.Result(compute)
	}

	constructor(json: JsonObject) : super(json) {
		val values = json["values"] as? JsonArray ?: throw JsonSyntaxException("Invalid `values` value: ${json["values"]}")

		this.values = values.stream().map {
			it as? JsonObject ?: throw JsonSyntaxException("JsonObject expected, $it <${it::class.qualifiedName}> given")

			when (val type = it["type"]?.asString?.lowercase() ?: throw JsonSyntaxException("Missing type")) {
				"tag" -> Tag(it)
				"key" -> Key(it)
				"constant" -> Constant(it)
				else -> throw JsonSyntaxException("Invalid type provided: $type")
			}
		}.collect(ImmutableList.toImmutableList())

		if (this.values.isEmpty()) {
			throw JsonSyntaxException("Compute action is empty")
		}
	}

	override fun toJson(): JsonObject {
		return super.toJson().also {
			it["action"] = "compute"
			it["values"] = JsonArray(values.size).also {
				for (value in values) {
					it.add(value.toJson())
				}
			}
		}
	}

	override fun update(registry: MutableCollection<MutableEntry>, modifier: ResourceLocation) {
		// do nothing
	}
}
